#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include <libnd.h>
#include <protocols/ip/libnd_ip.h>
#include "libnd_ipfrag.h"

/* Helper structures for the fragmentation -- this one describes
 * a specific fragment, through a start point in an IP packet's
 * payload and covering a given size.
 */
typedef struct lnd_frag_loc
{
  guint size;
  guint offset;
  gboolean more; /* Whether to set IP MF flag or not */
} LND_FragLoc;

/* The general fragmentation specification structure. It contains
 * a default fragment size and potentially a list of specific
 * LND_FragLoc fragment descriptions.
 */
struct lnd_frag_spec
{
  guint size; /* The default per-fragment size */
  GList *specials; /* Additional fragment requests, GList<LND_FragLoc*> */
  GList *indices; /* Indices of packets to fragment, data pointers are directly used as ints */
  guint num_fragged; /* Number of packets fragmented. */
};

/* A handle to the IP protocol spec, for convenience. */
static LND_Protocol *ipfrag_ip;
static int ipfrag_verbose;

/* Debugging macro that prints to stdout if verbose */
#define DV(x) do { if (ipfrag_verbose) { printf x ;} D(x); } while (0)

/* --- Private functions below --------------------------------------- */

static void
ipfrag_help(void)
{
  printf("This plugin allows flexible fragmentation of IP packets.\n"
	 "\n"
	 "USAGE: lndtool -r ipfrag \n"
	 "\n"
	 "  --help, -h, -?           This message.\n"
	 "  --fragment,-f <string>   Fragment format specifier. Its format is one or more\n"
         "                           entities of the form <size>@<offset><flags>, joined by commas,\n"
	 "                           with no whitespace inbetween. \"size\" is the payload size\n"
	 "                           of a fragment excluding the IP header, and offset is this\n"
	 "                           chunk's location inside an IP fragment's payload, where 0\n"
	 "                           is the first byte after the IP header. \"flags\" can either\n"
	 "                           be empty, in which case the fragment will not have the MF bit\n"
	 "                           set, or \"M\", in which case MF will be set.\n"
	 "\n"
	 "                           When \"offset\" is \"*\", offsets are fixed consecutively\n"
         "                           at 0, size, 2*size, etc, until the packet's IP payload is\n"
	 "                           exhausted. \"offset\", if given, must be multiple of 8. If\n"
	 "                           \"offset\" is \"*\", \"size\" must be a multiple of 8.\n"
	 "                           All but the last fragment created this way have MF set.\n"
	 "\n"
	 "  --input,-i <file>        Input trace file.\n"
	 "  --output,-o <file>       Output trace file.\n"
	 "\n"
	 "  --packet,-p <index>[,..] Indices of one or more packets to fragment, starting at 1.\n"
	 "                           If unspecified, all suitable packets will be fragmented.\n"
	 "                           The indices refer to the numbers in the original trace only.\n"
	 "\n"
	 "  --verbose,-v             Report what's going on in more detail.\n");
  
}

static LND_FragSpec *
ipfrag_spec_new(void)
{
  LND_FragSpec *spec;

  if (! (spec = g_new0(LND_FragSpec, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }

  return spec;
}

static void
ipfrag_spec_add_fragment(LND_FragSpec *spec, LND_FragLoc *loc)
{
  if (! spec)
    return;
  
  spec->specials = g_list_append(spec->specials, loc);
}

static LND_FragLoc *
ipfrag_fragloc_new(guint size, guint offset, gboolean more)
{
  LND_FragLoc *frag;

  if (! (frag = g_new0(LND_FragLoc, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }

  frag->offset = offset;
  frag->size = size;
  frag->more = more;

  return frag;
}


/* This function takes an IP packet, moves a subsection of its
 * payload forward to right after the IP header, and adjusts all
 * related size fields and checksums to shrink this packet to the
 * new content. If a packet doesn't contain the entire region
 * requested, the resulting packet is marked as a last fragment
 * (i.e., the IP MF bit is not set), otherwise as a fragment (i.e.,
 * the IP MF bit is set).
 */
static void
ipfrag_adjust_packet(LND_Packet *packet, guint offset, guint size, gboolean more)
{
  struct ip *iphdr;
  guchar *payload_start, *payload_end, *payload_ptr;
  guint base_size, effective_size = size;
  
  iphdr = (struct ip *) libnd_packet_get_data(packet, ipfrag_ip, 0);
  payload_start = payload_ptr = ((guchar *) iphdr) + (iphdr->ip_hl * 4);
  payload_end = libnd_packet_get_data_end(packet, ipfrag_ip, 0);
  base_size = payload_start - libnd_packet_get_data(packet, NULL, 0);
  
  /* If the requested offset is beyond what the packet provides,
   * we cannot do anything.
   */
  if (payload_start + offset > payload_end)
    return;
  
  /* Figure out effective size of payload we can shift: near the end
   * of the packet we may not have all the data demanded.
   */
  if (payload_start + offset + size > payload_end)
    effective_size = payload_end - (payload_start + offset);

  /* Adjust pcap-level packet size */
  packet->ph.caplen = base_size + effective_size;
  packet->ph.len = base_size + effective_size;
  
  /* Shift relevant payload forward to right after the header */
  memmove(payload_start, payload_start + offset, effective_size);
  
  /* Adjust IP fragment offset: */
  iphdr->ip_off = htons((ntohs(iphdr->ip_off) & ~IP_OFFMASK) | ((offset / 8) & IP_OFFMASK));

  /* Adjust IP MF flag as instructed: */
  if (more)
    iphdr->ip_off = htons(ntohs(iphdr->ip_off) | IP_MF);
  else
    iphdr->ip_off = htons(ntohs(iphdr->ip_off) & ~IP_MF);
    
  /* Adjust IP packet size -- the header plus the effective
   * payload size:
   */
  iphdr->ip_len = htons(iphdr->ip_hl * 4 + effective_size);

  /* Fix checksums: */
  libnd_packet_fix(packet);

  /* XXX: We could realloc the packet's data chunk here to actually get
   *      rid of all the data we no longer consider part of the packet.
   */

  /* Update protocol data structures as necessary: */
  libnd_packet_init(packet);
}

/*
 * This function iterates over all packets in the given trace, fragments
 * the IP packets according to the fragmentation specification in fs,
 * and writes all resulting fragments to the output dumper.
 */
static void
ipfrag_fragment_trace(LND_Trace *trace, LND_Dumper *dumper, LND_FragSpec *fs)
{
  int i = 1;
  GList *ind = fs->indices;
  LND_PacketIterator pit;
  LND_Packet *packet, *packet_orig, *tmp;
  
  for (libnd_pit_init_mode(&pit, trace, LND_PACKET_IT_AREA_R);
       libnd_pit_get(&pit);
       libnd_pit_next(&pit), i++)
    {
      packet_orig = libnd_pit_get(&pit);
      packet = NULL;
      
      if (! fs->indices || (ind && GPOINTER_TO_INT(ind->data) == i))
	{
	  packet = libnd_ipfrag_fragment_packet(packet_orig, fs);
	  ind = g_list_next(ind);
	}
      
      if (! packet)
	{
	  libnd_dumper_write(dumper, packet_orig);
	  continue;
	}
      
      while (packet)
	{
	  libnd_dumper_write(dumper, packet);
	  tmp = packet;
	  packet = packet->next;
	  libnd_packet_free(tmp);
	}
    }
}

static gint
ipfrag_int_cmp(gconstpointer a, gconstpointer b)
{
  if (GPOINTER_TO_INT(a) < GPOINTER_TO_INT(b))
    return -1;
  if (GPOINTER_TO_INT(a) > GPOINTER_TO_INT(b))
    return 1;
  
  return 0;
}

/* --- API calls below ----------------------------------------------- */

/* This function creates a LND_FragSpec structure out of a 
 * textual fragment specification.
 */
LND_FragSpec *
libnd_ipfrag_spec_new(const char *fragspec, const char *packets)
{
  LND_FragSpec *fs = NULL;
  const char *spec = fragspec;
  const char *specptr;
  char tmp[MAXPATHLEN];
  
  if (! fragspec || ! *fragspec)
    return NULL;
  
  if (! (fs = ipfrag_spec_new()))
    return NULL;
  
  /* Parse out format specifiers -- this is really not exactly nice: */
  while (spec)
    {
      guint offset = 0, size = 0;
      char flag = 0;

      specptr = strchr(spec, ',');
      
      if (! specptr)
	specptr = spec + strlen(spec);
      
      memcpy(tmp, spec, specptr-spec);
      tmp[specptr-spec] = '\0';
      
      if (sscanf(tmp, "%u@%u%c", &size, &offset, &flag) == 3 ||
	  sscanf(tmp, "%u@%u", &size, &offset) == 2)
	{
	  if (offset % 8 != 0)
	    {
	      fprintf(stderr, "%s: fragment offset must be multiple of 8, is %u, aborting.\n", tmp, offset);
	      libnd_ipfrag_spec_free(fs);
	      return NULL;
	    }

	  ipfrag_spec_add_fragment(fs, ipfrag_fragloc_new(size, offset, flag == 'M' ? TRUE : FALSE));
	}
      else if (sscanf(tmp, "%u@*", &size) == 1)
	{
	  if (size % 8 != 0)
	    {
	      fprintf(stderr, "%s: consecutive fragment size must be multiple of 8, is %u, aborting.\n", tmp, size);
	      libnd_ipfrag_spec_free(fs);
	      return NULL;
	    }
	  
	  fs->size = size;	    
	}
      else
	fprintf(stderr, "%s: format error in fragment specifier, skipping.\n", tmp);
      
      if (*specptr == '\0')
	break;
      
      spec = specptr + 1;
    }


  // Now create array with indices of packets to fragment.
  if (packets)
    {
      char *dup, *tmp;
      const char *comma = NULL;
      
      dup = tmp = strdup(packets);
      
      while ( (comma = strtok(tmp, ",")))
	{
	  fs->indices = g_list_append(fs->indices, GINT_TO_POINTER(atoi(comma)));
	  tmp = NULL;
	}
     
      // Sort packet indices.
      fs->indices = g_list_sort(fs->indices, ipfrag_int_cmp);
 
      g_free(dup);
    }
  
  return fs;
}

void
libnd_ipfrag_spec_free(LND_FragSpec *fs)
{
  GList *l;

  if (! fs)
    return;

  for (l = fs->specials; l; l = g_list_next(l))
    g_free(l->data);
  
  g_list_free(fs->specials);
  g_list_free(fs->indices);  
  g_free(fs);
}

LND_Packet *
libnd_ipfrag_fragment_packet(LND_Packet *packet, LND_FragSpec *fs)
{
  struct ip *iphdr;
  LND_Packet *p_frags = NULL, *p_frags_last = NULL;
  GList *specials;
  guchar *payload_start, *payload_end;
  guint offset = 0;
  gboolean fragmented = FALSE;

  if (! packet || ! fs)
    {
      D(("Input error.\n"));
      return NULL;
    }
  
  if (! (iphdr = (struct ip *) libnd_packet_get_data(packet, ipfrag_ip, 0)))
    {
      DV(("Packet %p doesn't contain IP\n", packet));
      return libnd_packet_duplicate(packet);
    }

  /* Don't fragment partially-captured packets.
   * XXX This could be done better by trying as much as we can
   *     despite partial captures.
   */
  if (packet->ph.caplen != packet->ph.len)
    {
      DV(("Not fragmenting partial capture.\n"));
      return libnd_packet_duplicate(packet);
    }
  
  /* Don't fragment existing fragments */
  if ((ntohs(iphdr->ip_off) & IP_OFFMASK) != 0 ||
      (ntohs(iphdr->ip_off) & IP_MF) != 0)
    {
      DV(("Not fragmenting fragment.\n"));
      return libnd_packet_duplicate(packet);
    }
  
  /* Sanity-check default fragment size */
  if (fs->size % 8 != 0)
    {
      DV(("Default fragment size (%u), must be multiple of 8.\n", fs->size));
      return libnd_packet_duplicate(packet);      
    }

  payload_start = ((guchar *) iphdr) + (iphdr->ip_hl * 4);
  payload_end = libnd_packet_get_data_end(packet, ipfrag_ip, 0);
  specials = fs->specials;
  
  /* Iterate over the payload and the specific fragmentation instructions,
   * create packets, and format them accordingly:
   */
  while ((fs->size > 0 && payload_start + offset < payload_end) || specials)
    {
      LND_Packet *p_copy = libnd_packet_duplicate(packet);
      LND_FragLoc *loc = NULL;

      if (! p_frags)
	p_frags = p_frags_last = p_copy;
      else
	{
	  p_frags_last->next = p_copy;
	  p_copy->prev = p_frags_last;
	  p_frags_last = p_copy;
	}

      /* If we have specific fragment requests, sanity-check them
       * and apply them whenever their start offset is before the
       * next regular offset.
       */
      if (specials)
	loc = (LND_FragLoc*) specials->data;
      
      /* If special fragment location is next, do that instead
       * of the regular chunk:
       */      
      if (loc && (fs->size == 0 || loc->offset <= offset))
	{
	  ipfrag_adjust_packet(p_copy, loc->offset, loc->size, loc->more);
	  specials = g_list_next(specials);
	}
      else
	{
	  ipfrag_adjust_packet(p_copy, offset, fs->size,
			       (payload_start + offset + fs->size < payload_end ? TRUE : FALSE));
	  offset += fs->size;
	}

      fragmented = TRUE;
    }
  
  fs->num_fragged += fragmented? 1 : 0;
  return p_frags;
}

/* --- Plugin hooks below -------------------------------------------- */

const char *name(void) { return "IPfrag"; }
const char *author(void) { return "Christian Kreibich"; }
const char *version(void) { return "0.1"; }

gboolean
init(void)
{
  if (! (ipfrag_ip = libnd_proto_registry_find(LND_PROTO_LAYER_NET,
					       0x0800)))
    return FALSE;
  
  return TRUE;
}

int
run(LND_Trace *trace, LND_PluginArgs *args)
{
  int i;
  const char *input = NULL, *output = NULL, *fragspec = NULL, *packets = NULL;
  LND_Dumper *dumper = NULL;
  LND_FragSpec *fs = NULL;

  for (i = 0; i < args->argc; i++)
    {
      if (!strcmp(args->argv[i], "-h")     ||
	  !strcmp(args->argv[i], "--help") ||
	  !strcmp(args->argv[i], "-?"))
	{
	  ipfrag_help();
	  return 0;
	}
      else if (!strcmp(args->argv[i], "-i") ||
	       !strcmp(args->argv[i], "--input"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass a trace file to the --input|-i option.\n");
	      return 1;
	    }
	  
	  input = args->argv[i];
	}
      else if (!strcmp(args->argv[i], "-o") ||
	       !strcmp(args->argv[i], "--output"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass a trace file to the --output|-o option.\n");
	      return 1;
	    }
	  
	  output = args->argv[i];
	}
      else if (!strcmp(args->argv[i], "-f") ||
	       !strcmp(args->argv[i], "--fragment"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass a fragmentation specifier to the --fragment|-f option.\n");
	      return 1;
	    }
	  
	  fragspec = args->argv[i];
	}
      else if (!strcmp(args->argv[i], "-p") ||
	       !strcmp(args->argv[i], "--packet"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass a packet index, starting at 1, to --packet|-p.\n");
	      return 1;
	    }
	  
	  packets = args->argv[i];
	}
      else if (!strcmp(args->argv[i], "-v") ||
	       !strcmp(args->argv[i], "--verbose"))
	{
	  ipfrag_verbose = TRUE;
	}
    }

  if (! input)
    fprintf(stderr, "Please provide an input file using --input|-i.\n");
  if (! output)
    fprintf(stderr, "Please provide an output file using --output|-o.\n");
  if (! fragspec)
    fprintf(stderr, "Please provide a fragmentation specification using --fragment|-f.\n");

  if (! input || ! output || ! fragspec)
    {
      ipfrag_help();
      return 1;
    }
  
  if (! (fs = libnd_ipfrag_spec_new(fragspec, packets)))
    return 1;
  
  if (! trace && ! (trace = libnd_trace_new(input)))
    {
      fprintf(stderr, "Could not open '%s'\n", input);
      return errno;
    }
  
  if (! (dumper = libnd_dumper_new(libnd_trace_get_pcap_handle(trace), output, NULL)))
    {
      fprintf(stderr, "Could not create output trace '%s'\n", output);
      return errno;
    }

  ipfrag_fragment_trace(trace, dumper, fs);

  DV(("Fragmented %u packets.\n", fs->num_fragged));

  libnd_ipfrag_spec_free(fs);
  libnd_dumper_free(dumper);
  libnd_trace_free(trace);
  
  return 0;
}
