#ifndef __libnd_ipfrag_h
#define __libnd_ipfrag_h

#include <libnd.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

typedef struct lnd_frag_spec LND_FragSpec;

/**
 * libnd_ipfrag_spec_new - creates a new fragmentation specification.
 * @fragspec: a textual fragmentation specification.
 * @packets: optional list of packet indices, of packets to fragment.
 *
 * The function takes a textual fragmentation specification and
 * constructs an internal datastructure for it. The resulting
 * pointer can then be passed to libnd_ipfrag_fragment_packet()
 * to describe how an IP packet is to be fragmented.
 *
 * The format of the specifier is one or more entities of the
 * form <size>@<offset><flags>, joined by commas with no whitespace
 * inbetween. "size" is the payload size of a fragment excluding
 * the IP header, and offset is this chunk's location inside an
 * IP fragment's payload, where 0 is the first byte after the IP
 * header. "flags" can either be empty, in which case the fragment
 * will not have the MF bit set, or "M", in which case MF will be
 * set.
 *
 * When "offset" is "*", offsets are fixed consecutively at 0,
 * size, 2*size, etc, until the packet's IP payload is exhausted.
 * "offset", if given, must be multiple of 8. If "offset" is "*",
 * "size" must be a multiple of 8. All but the last fragment created
 * this way have MF set.
 *
 * If string @packets is given, it must be of the form "<int>,<int>,..."
 * and contains a list of packets to fragment.
 *
 * Release the resulting structure using libnd_ipfrag_spec_free().
 */
LND_FragSpec *libnd_ipfrag_spec_new(const char *fragspec, const char *packets);

/**
 * libnd_ipfrag_spec_free - releases a fragmentation specification.
 * @fs: the specification to free.
 *
 * The function releases all memory associated with fs.
 */
void          libnd_ipfrag_spec_free(LND_FragSpec *fs);

/**
 * libnd_ipfrag_fragment_packet - fragments an IP packet.
 * @packet: the packet to fragment.
 * @fs: instructions on how to fragment.
 *
 * The function fragments @packet according to @fs's instructions.
 * The resulting fragments are returned as a linked sequence of
 * packets, with a pointer to the first packet. You need to ensure
 * these packets are cleaned up properly, either by hooking them
 * into a trace part or by deleting them directly, via libnd_packet_free().
 */
LND_Packet   *libnd_ipfrag_fragment_packet(LND_Packet *packet, LND_FragSpec *fs);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
