/*

Copyright (C) 2004 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
%{
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <libnd.h>
#include <libnd_phdl.h>
#include <libnd_phdl_exp.h>
#include <libnd_phdl_reg.h>
#include <libnd_phdl_header.h>

int phdllex(void);
int phdlparse(void);
int phdlerror(char *, ...);

#define yylex phdllex
#define yyparse phdlparse
#define yyerror phdlerror

int parse_errors = 0;
int parse_lineno = 0;
const char *parse_filename = NULL;

%}

%union {
  char                   *string;
  gint64                  number;
  LND_PHDL_LayerDef      *layer;
  LND_PHDL_NumParam      *val;
  LND_PHDL_NumExp        *numexp;
  LND_PHDL_BoolExp       *boolexp;
  LND_PHDL_FieldParams   *params;
  LND_PHDL_ChainEnd      *chainend;
  LND_PHDL_FieldTempl    *ft;
  LND_PHDL_HeaderTempl   *ht;
}

%token PROTO DEF BLOCK HEX INT STR ENUM UNIT SCALE SEP CHAIN UNION
%token IF OTHER EQU NEQ LET GET LEQ GEQ TRU FLS LINK NET TRANS APP
%token VAL UNTIL ITEMS SEEN LENGTH UNSIGNED
%token <string> STRING
%token <string> FIELD
%token <number> NUMBER
%token <number> HEXNUMBER
%type <layer> layerdef
%type <val> val
%type <params> params
%type <numexp> numexp
%type <boolexp> boolexp
%type <number> demuxmark unsignedmark
%type <chainend> chainend
%type <number> numval
%type <number> unioncontent
%type <ft> intfield hexfield strfield enumfield structfield structdef
%type <ft> block chain union groupfield atomicfield content
%type <ht> headerdef

%left '+' '-'
%left '*' '/'
%left '%'

%%

statement       : /* empty */
                | statement structdef
                | statement headerdef
                ;

structdef       : DEF STRING '{'
	{
		LND_PHDL_FieldTempl *st;

		D(("Creating new structure: '%s'\n", $2));
		st = libnd_phdl_ft_struct_new($2);
		libnd_phdl_ctxt_push(st);
		g_free($2);
	}
		  contentlist '}'
	{
		LND_PHDL_FieldTempl *st = libnd_phdl_ctxt_pop();
		
		libnd_phdl_structreg_add(st);
		$$ = st;
	}
		;

headerdef	: PROTO STRING '(' layerdef ')' '{' 
	{
		LND_PHDL_HeaderTempl *ht;
		LND_PHDL_LayerDef *ld = $4;
	
		D(("Creating new header: '%s'\n", $2));
		ht = libnd_phdl_ht_new($2, ld->layer, ld->magic);
		g_free(ld);
		g_free($2);

		libnd_phdl_ht_ctxt_set(ht);
		libnd_phdl_ctxt_push(ht->root);
	}
		  contentlist '}'
	{
		LND_PHDL_HeaderTempl *ht = libnd_phdl_ht_ctxt_get();

		(void) libnd_phdl_ctxt_pop();
		libnd_phdl_htreg_add(ht);
		D(("Header '%s' registered - done!\n", ht->name));
	}
		;

contentlist	: /* empty */
		| contentlist content sep
		;

content         : groupfield
		| atomicfield
		;

sep		: /* empty */
                | ';'
		;

groupfield      : block
                | union
		| enumfield
		| chain
		;

atomicfield     : intfield
		| hexfield
		| strfield
                | structfield
		;

structfield     : FIELD STRING
	{
		LND_PHDL_FieldTempl *ft = libnd_phdl_structreg_get($1);

		if (!ft)
		  yyerror("Undefined structure '%s'", $1);
	
		$$ = libnd_phdl_ft_clone(ft, $2);
		g_free($1); g_free($2);
	}
		;

demuxmark	: /* empty */
	{	$$ = FALSE; }
		| '!'
	{	$$ = TRUE; D(("DEMUXMARK\n")); }
		;

unsignedmark	: /* empty */
	{	$$ = FALSE; }
		| UNSIGNED
	{	$$ = TRUE; D(("UNSIGNED\n")); }
		;

layerdef	: LINK  ':' numval
	{
		LND_PHDL_LayerDef *ld;

		D(("link layer magic: 0x%lx\n", (long unsigned) $3));
		ld = g_new0(LND_PHDL_LayerDef,1);
		ld->layer = LND_PROTO_LAYER_LINK;
		ld->magic = $3;
		$$ = ld;
	}
		| NET   ':' numval
	{
		LND_PHDL_LayerDef *ld;

		D(("net layer magic: 0x%lx\n", (long unsigned) $3));
		ld = g_new0(LND_PHDL_LayerDef,1);
		ld->layer = LND_PROTO_LAYER_NET;
		ld->magic = $3;
		$$ = ld;
	}
		| TRANS ':' numval
	{
		LND_PHDL_LayerDef *ld;

		D(("trans layer magic: 0x%lx\n", (long unsigned) $3));
		ld = g_new0(LND_PHDL_LayerDef,1);
		ld->layer = LND_PROTO_LAYER_TRANS;
		ld->magic = $3;
		$$ = ld;
	}
		| APP   ':' numval
	{
		LND_PHDL_LayerDef *ld;

		D(("app layer magic: %u\n", (unsigned) $3));
		ld = g_new0(LND_PHDL_LayerDef,1);
		ld->layer = LND_PROTO_LAYER_APP;
		ld->magic = $3;
		$$ = ld;
	}
		;


block		: BLOCK STRING '{'
	{
		LND_PHDL_FieldTempl *block;
		
		D(("New block: '%s'\n", $2));
		block = libnd_phdl_ft_block_new($2);
		libnd_phdl_ctxt_push(block);		
		g_free($2);
	}
		  contentlist '}'
	{
		LND_PHDL_FieldTempl *block = libnd_phdl_ctxt_pop();
		D(("Finishing block '%s'\n", block->name));
		$$ = block;
	}
		;

intfield	: demuxmark unsignedmark INT STRING numexp params
	{
		LND_PHDL_FieldTempl  *intfield;
                LND_PHDL_FieldParams *params = $6;

		D(("New intfield: '%s', width %s\n", $4,
		   libnd_phdl_numexp_to_string($5)));
		intfield = libnd_phdl_ft_int_new($4, $5);
		g_free($4);

		if ($1 == TRUE)
		  {
		    LND_PHDL_HeaderTempl *ht = libnd_phdl_ht_ctxt_get();
		    ht->demux_field = intfield;
		  }

                if (params)
                  {
                    if (! libnd_phdl_ft_set_params(intfield, params))
                      yyerror("Cannot set parameters on element %s", intfield->name);

                    libnd_phdl_ft_fieldparams_free(params);
                  }

		$$ = intfield;
	}
		;

hexfield	: demuxmark unsignedmark HEX STRING numexp params
	{
		LND_PHDL_FieldTempl  *hexfield;
                LND_PHDL_FieldParams *params = $6;

		D(("New hexfield: '%s', width %s\n", $4,
		   libnd_phdl_numexp_to_string($5)));
		hexfield = libnd_phdl_ft_hex_new($4, $5);
		g_free($4);

		if ($1 == TRUE)
		  {
		    LND_PHDL_HeaderTempl *ht = libnd_phdl_ht_ctxt_get();
		    ht->demux_field = hexfield;
		  }

                if (params)
                  {
                    if (! libnd_phdl_ft_set_params(hexfield, params))
                      yyerror("Cannot set parameters on element %s", hexfield->name);

                    libnd_phdl_ft_fieldparams_free(params);
                  }

		$$ = hexfield;
	}
		;

strfield	: demuxmark STR STRING numexp params
	{
		LND_PHDL_FieldTempl  *strfield;
                LND_PHDL_FieldParams *params = $5;

		D(("New string item: '%s'\n", $3));
		strfield = libnd_phdl_ft_str_fixed_new($3, $4);
		g_free($3);

		if ($1 == TRUE)
		  {
		    LND_PHDL_HeaderTempl *ht = libnd_phdl_ht_ctxt_get();
		    ht->demux_field = strfield;
		  }

                if (params)
                  {
                    if (! libnd_phdl_ft_set_params(strfield, params))
                      yyerror("Cannot set parameters on element %s", strfield->name);

                    libnd_phdl_ft_fieldparams_free(params);
                  }

		$$ = strfield;
	}
		| demuxmark STR STRING STRING params
	{
		LND_PHDL_FieldTempl  *strfield;
                LND_PHDL_FieldParams *params = $5;

		D(("New string item: '%s', terminated by %s.\n", $3, $4));
		strfield = libnd_phdl_ft_str_variable_new($3, $4);
		g_free($3);
		g_free($4);

		if ($1 == TRUE)
		  {
		    LND_PHDL_HeaderTempl *ht = libnd_phdl_ht_ctxt_get();
		    ht->demux_field = strfield;
		  }

                if (params)
                  {
                    if (! libnd_phdl_ft_set_params(strfield, params))
                      yyerror("Cannot set parameters on element %s", strfield->name);

                    libnd_phdl_ft_fieldparams_free(params);
                  }

		$$ = strfield;
	}
		;

enumfield	: demuxmark ENUM STRING numexp '{'
	{
		LND_PHDL_FieldTempl *en;
		
		D(("New enumfield: '%s'\n", $3));
		en = libnd_phdl_ft_enum_new($3, $4);

		libnd_phdl_ctxt_push(en);
		g_free($3);

		if ($1 == TRUE)
		  {
		    LND_PHDL_HeaderTempl *ht = libnd_phdl_ht_ctxt_get();
		    ht->demux_field = en;
		  }
	}
		  enumitems '}'
	{
		$$ = libnd_phdl_ctxt_pop();
	}
		;
enumitems	: /* empty */
		| enumitems HEXNUMBER ':' STRING ';'
	{
		LND_PHDL_FieldTempl *en = libnd_phdl_ctxt_get();
		LND_PHDL_EnumItem *item;
	
		D(("New hex enum item: '%s' for %i\n", $4, (int) $2));
		libnd_phdl_ft_enum_set_hex_format(en, TRUE);
		item = libnd_phdl_ft_enumitem_new($4, $2);
		libnd_phdl_ft_enum_add_item(en, item);
		g_free($4);
	}
		| enumitems NUMBER ':' STRING ';'
	{
		LND_PHDL_FieldTempl *en = libnd_phdl_ctxt_get();
		LND_PHDL_EnumItem *item;
	
		D(("New enum item: '%s' for %i\n", $4, (int) $2));
		item = libnd_phdl_ft_enumitem_new($4, $2);
		libnd_phdl_ft_enum_add_item(en, item);
		g_free($4);
	}
		;

params  	: /* empty */
	{
		$$ = NULL;
	}
		| '{'
	{
		LND_PHDL_FieldParams *params = libnd_phdl_ft_fieldparams_new();

		D_ASSERT_PTR(params);
		libnd_phdl_ctxt_push(params);		
	}
		  paramelems '}'
	{		
		$$ = libnd_phdl_ctxt_pop();
	}
		;

paramelems 	: /* empty */
                | paramelems UNIT  '=' numval ';'
	{
		LND_PHDL_FieldParams *params = libnd_phdl_ctxt_get();

		params->unit = $4;
		D(("Parsed output manipulator UNIT = %i\n", (int) $4));
	}
		| paramelems SCALE '=' numval ';'
	{
		LND_PHDL_FieldParams *params = libnd_phdl_ctxt_get();

		params->scale = $4;
		D(("Parsed output manipulator SCALE = %i\n", (int) $4));
	}
		| paramelems SEP   '=' STRING ';'
	{
		LND_PHDL_FieldParams *params = libnd_phdl_ctxt_get();

		params->sep = $4;
		D(("Parsed output manipulator SEP = '%s'\n", $4));
	}
		| paramelems VAL   '=' STRING ';'
	{
		LND_PHDL_FieldParams *params = libnd_phdl_ctxt_get();

		params->val.str_val = $4;
		D(("Parsed output option VAL = '%s'\n", $4));
	}
		;

union		: UNION STRING '{'
	{
		LND_PHDL_FieldTempl *un;
		
		D(("New union: '%s'\n", $2));
		un = libnd_phdl_ft_union_new($2);
		libnd_phdl_ctxt_push(un);
		g_free($2);
	}
		  unioncontent '}'
	{
		LND_PHDL_FieldTempl *un = libnd_phdl_ctxt_pop();
		D(("Finishing union '%s'\n", un->name));
		$$ = un;
	}
		;
unioncontent	: /* empty */
	{
	}
		| unioncontent content params IF '(' boolexp ')' ';'
        {
		LND_PHDL_FieldTempl  *un = libnd_phdl_ctxt_get();

		libnd_phdl_ft_set_params($2, $3);
		libnd_phdl_ft_union_add_field(un, $2, $6);
		libnd_phdl_ft_fieldparams_free($3);
	}
		;

chain		: CHAIN STRING '{' 
	{
		LND_PHDL_FieldTempl *chain;
		
		D(("New chain: '%s'\n", $2));
		chain = libnd_phdl_ft_chain_new($2);
		libnd_phdl_ctxt_push(chain);
		g_free($2);
	}
		  contentlist '}' chainend
	{
		LND_PHDL_FieldTempl *chain = libnd_phdl_ctxt_pop();
		LND_PHDL_ChainEnd *chain_end = $7;

		D(("Finishing chain '%s'\n", chain->name));
		libnd_phdl_ft_chain_set_end(chain, chain_end);
		$$ = chain;
	}
		;

chainend	: UNTIL ITEMS numexp
	{
		$$ = libnd_phdl_ft_chainend_new_items($3);
	}
		| UNTIL SEEN FIELD '{'
        {
		LND_PHDL_FieldTempl *templ;
		LND_PHDL_ChainEnd *end;

		if (! (templ = libnd_phdl_structreg_get($3)))
		  yyerror("Field '%s' not defined", $3);

		end = libnd_phdl_ft_chainend_new_pattern(templ);
		libnd_phdl_ctxt_push(end);
	}
		  chainendpatterns '}'
	{
		$$ = libnd_phdl_ctxt_pop();
	}
		| UNTIL LENGTH numexp
	{
		$$ = libnd_phdl_ft_chainend_new_length($3);
	}
		;

chainendpatterns : /* empty */
		| chainendpatterns boolexp ';'
	{
		LND_PHDL_ChainEnd *end = libnd_phdl_ctxt_get();

		end->fieldexps = g_list_append(end->fieldexps, $2);
	}
		;

boolexp		: val EQU val
	{
		$$ = libnd_phdl_boolexp_new($1, $3, LND_PHDL_EQU);
	}
		| val NEQ val
	{
		$$ = libnd_phdl_boolexp_new($1, $3, LND_PHDL_NEQ);
	}
		| val LET val
	{
		$$ = libnd_phdl_boolexp_new($1, $3, LND_PHDL_LET);
	}
		| val GET val
	{
		$$ = libnd_phdl_boolexp_new($1, $3, LND_PHDL_GET);
	}
		| val LEQ val
	{
		$$ = libnd_phdl_boolexp_new($1, $3, LND_PHDL_LEQ);
	}
		| val GEQ val
	{
		$$ = libnd_phdl_boolexp_new($1, $3, LND_PHDL_GEQ);
	}
		| TRU
	{
		$$ = libnd_phdl_boolexp_new(libnd_phdl_numparam_lit_new(TRUE), NULL, LND_PHDL_NOB);
	}
		| FLS
	{
		$$ = libnd_phdl_boolexp_new(libnd_phdl_numparam_lit_new(FALSE), NULL, LND_PHDL_NOB);
	}
		| '(' boolexp ')'
	{
		$$ = $2;
	}
		;

numexp		: val
	{
		$$ = libnd_phdl_numexp_new($1, NULL, LND_PHDL_NON);
	}
		| numexp '+' numexp
	{
		$$ = libnd_phdl_numexp_new(libnd_phdl_numparam_exp_new($1),
                                           libnd_phdl_numparam_exp_new($3),
                                           LND_PHDL_ADD);
	}
		| numexp '-' numexp
	{
		$$ = libnd_phdl_numexp_new(libnd_phdl_numparam_exp_new($1),
                                           libnd_phdl_numparam_exp_new($3),
                                           LND_PHDL_SUB);
	}
		| numexp '/' numexp
	{
		$$ = libnd_phdl_numexp_new(libnd_phdl_numparam_exp_new($1),
                                           libnd_phdl_numparam_exp_new($3),
                                           LND_PHDL_DIV);
	}
		| numexp '*' numexp
	{
		$$ = libnd_phdl_numexp_new(libnd_phdl_numparam_exp_new($1),
                                           libnd_phdl_numparam_exp_new($3),
                                           LND_PHDL_MUL);
	}
		| numexp '%' numexp
	{
		$$ = libnd_phdl_numexp_new(libnd_phdl_numparam_exp_new($1),
                                           libnd_phdl_numparam_exp_new($3),
                                           LND_PHDL_MOD);
	}
		| '(' numexp ')'
	{
		$$ = $2;
	}
		;
val		: numval
	{
	        $$ = libnd_phdl_numparam_lit_new($1);
	}
		| FIELD
	{
		$$ = libnd_phdl_numparam_field_new($1);
	}
		;

numval		: NUMBER
	{
	        $$ = $1;
	}
                | HEXNUMBER
	{
	        $$ = $1;
	}
		;
%%

int
yyerror(char *fmt, ...)
{
	va_list ap;
	parse_errors = 1;

	va_start(ap, fmt);
	fprintf(stderr, "%s:%d: ", parse_filename, parse_lineno);
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
	return (0);
}


int
libnd_phdl_parse_template(const char *filename)
{
	extern FILE *phdlin;

	if (! (phdlin = fopen(filename, "r")))
	  {
	    D(("Error opening template file %s.\n", filename));
	    return -1;
          }

	parse_lineno = 1;
	parse_filename = filename;

	yyparse();

	fclose(phdlin);

	return (parse_errors ? -1 : 0);
}
